#pypp 0
// Iris: micro-kernel for a capability-based operating system.
// schedule.ccp: Thread scheduling.
// Copyright 2009 Bas Wijnen <wijnen@debian.org>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include "kernel.hh"

static void run_thread (Thread *thread):
	thread->schedule_next = first_scheduled
	if thread->schedule_next:
		thread->schedule_next->schedule_prev = thread
	first_scheduled = thread

static void unrun_thread (Thread *thread):
	if current == thread:
		current = thread->schedule_next
	if thread->schedule_prev:
		thread->schedule_prev->schedule_next = thread->schedule_next
	else:
		first_scheduled = thread->schedule_next
	if thread->schedule_next:
		thread->schedule_next->schedule_prev = thread->schedule_prev

void Thread::run ():
	if flags & THREAD_FLAG_RUNNING:
		return
	flags |= THREAD_FLAG_RUNNING
	if flags & THREAD_FLAG_WAITING:
		return
	run_thread (this)

void Thread::unrun ():
	if !(flags & THREAD_FLAG_RUNNING):
		return
	flags &= ~THREAD_FLAG_RUNNING
	if flags & THREAD_FLAG_WAITING:
		return
	unrun_thread (this)

void Thread::wait ():
	if flags & THREAD_FLAG_WAITING:
		return
	if flags & THREAD_FLAG_RUNNING:
		unrun_thread (this)
	flags |= THREAD_FLAG_WAITING
	// Try to receive a message from a Receiver immediately.
	for Receiver *r = receivers; r; r = r->next_owned:
		if r->try_deliver ():
			break

void Thread::unwait ():
	if !(flags & THREAD_FLAG_WAITING):
		return
	flags &= ~THREAD_FLAG_WAITING
	if flags & THREAD_FLAG_RUNNING:
		run_thread (this)

void schedule ():
	Thread *old = current
	if current:
		current = current->schedule_next
	if !current:
		current = first_scheduled
